﻿local Links = LibStub("AceAddon-3.0"):NewAddon("Links", "AceConsole-3.0", "AceHook-3.0", "AceEvent-3.0", "AceConsole-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale("Links")

local options = {
	desc = L["Links - plenty of them"],
	type = "group",
	childGroups = "tab",
	args = {
		options = {
			order = 1,
			name = L["Options"],
			type = "group",
			args = {
				modinvert = {
					width = "full",
					name = L["modinvert"],
					desc = L["Invert the alt-clic modifier behavior of Links."],
					type = "toggle",
					get = function() return Links.db.profile.modinvert end,
					set = function(_, value) Links.db.profile.modinvert = value end,
				},
				autohide = {
					width = "full",
					name = L["autohide"],
					desc = L["Automatically hide tooltip on escape."],
					type = "toggle",
					get = function() return Links.db.profile.autohide end,
					set = function(_, value) Links:SetAutoHide(value) end,
				},
				autocomp = {
					width = "full",
					name = L["autocomp"],
					desc = L["Automatically show the compare window when hovering tooltips."],
					type = "toggle",
					get = function() return Links.db.profile.autocomp end,
					set = function(_, value) Links.db.profile.autocomp = value end,
				},
				showicon = {
					width = "full",
					name = L["showicon"],
					desc = L["Show item icon with the tooltip."],
					type = "toggle",
					get = function() return Links.db.profile.showicon end,
					set = function(_, value) Links.db.profile.showicon = value end,
				},
				history = {
					width = "full",
					name = L["Link history"],
					desc = L["Keep history of opened links."],
					type = "toggle",
					get = function() return Links.db.profile.history end,
					set = function(_, value)
						Links.db.profile.history = value
						Links.menu = nil
						Links.linkHistory = nil
					end,
				},
				types = {
					width = "full",
					name = L["Link types"],
					type = "multiselect",
					values = {
						unit = L["Unit"],
						item = L["Item"],
						quest = L["Quest"],
						spell = L["Spell"],
						enchant = L["Enchantments and Crafts"],
						achievement = ACHIEVEMENTS,
						talent = TALENTS,
					},
					get = function (_, type) return Links.db.profile.types[type] end,
					set = function (_, type, value) Links.db.profile.types[type] = value end,
				}
			},
		},
	},
}

local defaults = {
	profile = {
		modinvert = false,
		autohide = true,
		autocomp = false,
		showicon = false,
		history = false,
		maxhistory = 25,
		positions = {},
		types = {
			unit = false,
			item = true,
			spell = true,
			enchant = true,
			quest = false,
			achievement = true,
			talent = false,
			-- player = nil,
			-- BNplayer = nil,
		}
	}
}

function Links:SavePosition(frame)
	local index = frame.index
	local position = self.db.profile.positions[index]
	if not position then
		position = {}
		self.db.profile.positions[index] = position
	end
	position.left, position.top = frame:GetLeft(), frame:GetTop()
end

function Links:OnInitialize()
	self.frames = {}
	self.index = 1

	self.db = LibStub("AceDB-3.0"):New("LinksDB", defaults)

	local AceDBOptions = LibStub("AceDBOptions-3.0", true)
	if AceDBOptions then
		options.args.profiles = AceDBOptions:GetOptionsTable(self.db)
		options.args.profiles.order = 200
	end

	LibStub("AceConfig-3.0"):RegisterOptionsTable("Links", options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Links", "Links")
	self:RegisterChatCommand("links", "OnChatCommand")
end

function Links:OnChatCommand(input)
	if not input or input:trim() == "" then
		InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
	else
		LibStub("AceConfigCmd-3.0").HandleCommand(self, "links", "Links", input)
	end
end

Links.functionNames = {}
Links.functionHandlers = {}
Links.shoppingTooltips = { ShoppingTooltip1, ShoppingTooltip2, ShoppingTooltip3 }

function Links:AddToMenu(name, handler)
	local names, handlers = self.functionNames, self.functionHandlers
	names[#names + 1] = name
	handlers[#handlers + 1] = function (_, frame) handler(frame, frame.text, frame.link) end
end

Links:AddToMenu(L["Link in chat"], function (_, text)
	if not ChatFrameEditBox:IsVisible() then
		ChatFrame_OpenChat (text)
	else
		ChatFrameEditBox:Insert(text)
	end
	Links.menu = nil
end)

Links:AddToMenu(L["Dressup"], function (_, _, link)
	DressUpItemLink(link)
end)

Links:AddToMenu(L["Close all Tooltips"], function ()
	Links:HideAllFrames()
end)

Links:AddToMenu(L["Options"], function ()
	InterfaceOptionsFrame_OpenToCategory(Links.optionsFrame)
end)

function Links:OnEnable()
	-- As long as CombatLog_LoadUI hooks SetItemRef,
	-- we must make sure it's done before we do.
	CombatLog_LoadUI()
	self:RawHook("SetItemRef", true)
end

function Links:SetAutoHide(value)
	local UISpecialFrames = UISpecialFrames
	self.db.profile.autohide = value
	for frame in pairs(self.frames) do
		local name = frame:GetName()
		if value then
			UISpecialFrames[#UISpecialFrames + 1] = name
		else
			for i = 1, #UISpecialFrames do
				if name == UISpecialFrames[i] then
					table.remove(UISpecialFrames, i)
					break
				end
			end
		end
	end
end

function Links:MakeMenu(frame)
	local menu = self.menu
	if not menu then
		menu = {}
		if self.db.profile.history then
			if not self.linkHistory then
				self.linkHistory = {}
			end
			menu[#menu + 1] = {
				text = L["Link history"],
				hasArrow = true,
				menuList = self.linkHistory
			}
		end
		for i, name in ipairs(self.functionNames) do
			menu[#menu + 1] = {
				text = name,
				func = self.functionHandlers[i],
				arg1 = frame,
			}
		end
		self.menu = menu
	else
		for _, entry in ipairs(menu) do
			entry.arg1 = frame
		end
	end
	return menu
end

function Links:OpenMenu(frame)
	local dropdown = self.dropdown
	if not dropdown then
		dropdown = CreateFrame("Frame", "LinksDropDown", nil, "UIDropDownMenuTemplate")
		dropdown.point = "TOPLEFT"
		dropdown.relativePoint = "TOPRIGHT"
		dropdown.displayMode = "MENU"
		self.dropdown = dropdown
	end
	dropdown.relativeTo = frame

	EasyMenu(self:MakeMenu(frame), dropdown)
end

function Links:HideAllFrames()
	for frame in pairs(self.frames) do
		if frame:IsVisible() then
			frame:Hide()
		end
	end
end

local LinksFrame_Events = {
	OnDragStart = function (self)
		Links:HideCompareItems(self)
		self:StartMoving()
	end,
	OnDragStop = function (self)
		self:StopMovingOrSizing()
		Links:ValidateFramePosition(self)
		if Links.db.profile.autocomp or IsModifiedClick("COMPAREITEMS") then
			Links:ShowCompareItem(self)
		end
	end,
	OnHide = function (self)
		local left, top = self:GetLeft(), self:GetTop()
		self:ClearAllPoints()
		self:SetPoint("TOPLEFT", left, top - UIParent:GetHeight())
		self.link = nil
		self.text = nil
		Links:SavePosition(self)
	end,
	OnReceiveDrag = function (self)
		local type, id, link = GetCursorInfo()
		if type == "item" then
			Links:OpenLink(nil, link, self)
			ClearCursor()
		end
	end,
	OnEnter = function (self)
		local autocomp = Links.db.profile.autocomp
		if autocomp or IsModifiedClick("COMPAREITEMS") then
			Links:ShowCompareItem(self)
		end
		if not autocomp then
			self:RegisterEvent("MODIFIER_STATE_CHANGED")
		end
	end,
	OnLeave = function (self)
		Links:HideCompareItems(self)
		self:UnregisterEvent("MODIFIER_STATE_CHANGED")
	end,
	OnEvent = function (self, event)
		if event == "MODIFIER_STATE_CHANGED" then
			if IsModifiedClick("COMPAREITEMS") then
				Links:ShowCompareItem(self)
			else
				Links:HideCompareItems(self)
			end
		end
	end,
	OnTooltipSetItem = function (...)
		local handler = ItemRefTooltip:GetScript"OnTooltipSetItem"
		return handler and handler(...)
	end,
	OnTooltipCleared = function (...)
		local handler = ItemRefTooltip:GetScript"OnTooltipCleared"
		return handler and handler(...)
	end,
}

function Links:ShowCompareItem(frame, shift)
	local _, link = frame:GetItem()
	if not link then return end

	local shoppingTooltip1, shoppingTooltip2, shoppingTooltip3 = unpack(frame.shoppingTooltips)

	local item1 = shoppingTooltip1:SetHyperlinkCompareItem(link, 1, shift, frame)
	local item2 = shoppingTooltip2:SetHyperlinkCompareItem(link, 2, shift, frame)
	local item3 = shoppingTooltip3:SetHyperlinkCompareItem(link, 3, shift, frame)

	local leftPos = frame:GetLeft() or 0
	local rightPos = frame:GetRight() or 0
	local screenWidth = GetScreenWidth()
	local rightDist = screenWidth - rightPos

	local side_left = (rightDist < leftPos)

	local anchorType = frame:GetAnchorType()
	if anchorType and anchorType ~= "ANCHOR_PRESERVE" then
		local totalWidth = item1 and shoppingTooltip1:GetWidth() or 0
		if item2 then
			totalWidth = totalWidth + shoppingTooltip2:GetWidth()
		end
		if item3 then
			totalWidth = totalWidth + shoppingTooltip3:GetWidth()
		end

		if side_left then
			if totalWidth > leftPos then
				frame:SetAnchorType(anchorType, (totalWidth - leftPos), 0)
			end
		else
			totalWidth = totalWidth + rightPos
			if totalWidth >  screenWidth then
				frame:SetAnchorType(anchorType, -(totalWidth - screenWidth), 0)
			end
		end
	end

	local anchorFrom, anchorTo
	if side_left then
		anchorFrom, anchorTo = "TOPRIGHT", "TOPLEFT"
	else
		anchorFrom, anchorTo = "TOPLEFT", "TOPRIGHT"
	end
	local owner = frame
	if item3 then
		shoppingTooltip3:SetOwner(owner, "ANCHOR_NONE")
		shoppingTooltip3:ClearAllPoints()
		shoppingTooltip3:SetPoint(anchorFrom, owner, anchorTo, 0, -10)
		shoppingTooltip3:SetHyperlinkCompareItem(link, 3, shift, frame)
		shoppingTooltip3:Show()
		owner = shoppingTooltip3
	end
	if item1 then
		shoppingTooltip1:SetOwner(owner, "ANCHOR_NONE")
		shoppingTooltip1:ClearAllPoints()
		shoppingTooltip1:SetPoint(anchorFrom, owner, anchorTo)
		shoppingTooltip1:SetHyperlinkCompareItem(link, 1, shift, frame)
		shoppingTooltip1:Show()

		if item2 then
			shoppingTooltip2:SetOwner(shoppingTooltip1, "ANCHOR_NONE")
			shoppingTooltip2:ClearAllPoints()
			shoppingTooltip2:SetPoint(anchorFrom, shoppingTooltip1, anchorTo)
			shoppingTooltip2:SetHyperlinkCompareItem(link, 2, shift, frame)
			shoppingTooltip2:Show()
		end
	end
end

function Links:HideCompareItems(frame)
	local shoppingTooltip1, shoppingTooltip2, shoppingTooltip3 = unpack(frame.shoppingTooltips)
	if shoppingTooltip1:IsOwned(frame) then
		shoppingTooltip1:Hide()
		shoppingTooltip2:Hide()
		shoppingTooltip3:Hide()
	end
end

Links.RegisterHookFrame = function()
	print("Deprecated API Links:RegisterHookFrame() called")
end

local function CreateTexture (parent, path, blend)
	local t = parent:CreateTexture()
	t:SetAllPoints(parent)
	t:SetTexture(path)
	if blend then
		t:SetBlendMode(blend)
	end
	return t
end

function Links:MakeFrame()
	local name = string.format("LinksTooltip%d", self.index)
	local frame = CreateFrame("GameTooltip", name, UIParent, "GameTooltipTemplate")
	frame.index = self.index

	self.index = self.index + 1
	self.frames[frame] = true

	local close = CreateFrame("Button", name.."Close", frame)
	frame.close = close


	close:SetNormalTexture(CreateTexture(close, "Interface\\Buttons\\UI-Panel-MinimizeButton-Up"))
	close:SetPushedTexture(CreateTexture(close, "Interface\\Buttons\\UI-Panel-MinimizeButton-Down"))
	close:SetHighlightTexture(CreateTexture(close, "Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight", "ADD"))

	close:RegisterForClicks("LeftButtonUp", "RightButtonUp")
	close:SetScript("OnClick",
		function (_, button)
			if button == "LeftButton" then
				frame:Hide()
			else
				self:OpenMenu(frame)
			end
		end)


	close:SetWidth(32)
	close:SetHeight(32)
	close:ClearAllPoints()
	close:SetPoint("TOPRIGHT", 1, 0)

	close:Show()

	frame:EnableMouse(true)
	frame:SetMovable(true)
	frame:SetToplevel(true)
	frame:SetFrameStrata("HIGH")
	local position = self.db.profile.positions[frame.index]
	if position then
		frame:SetPoint("TOPLEFT", position.left, position.top - UIParent:GetHeight())
	else
		frame:SetPoint("CENTER")
	end

	frame:SetPadding(16)
	frame:RegisterForDrag("LeftButton")

	for event, handler in pairs(LinksFrame_Events) do
		frame:SetScript(event, handler)
	end

	if self.db.profile.autohide then
		UISpecialFrames[#UISpecialFrames + 1] = name
	end

	frame.shoppingTooltips = self.shoppingTooltips
	return frame
end

function Links:ValidateFramePosition(frame)
	local left, top = frame:GetLeft(), frame:GetTop()
	frame:ClearAllPoints()
	frame:SetPoint("TOPLEFT", left, top-UIParent:GetHeight())
	ValidateFramePosition(frame)
	self:SavePosition(frame)
end

function Links:GetAvailableFrame()
	for frame in pairs(self.frames) do
		if not frame:IsVisible() then
			return frame
		end
	end
	return self:MakeFrame()
end

do
	local event
	function Links:TriggerEvent(link, frame)
		if event == nil then
			event = AceLibrary and AceLibrary:HasInstance("AceEvent-2.0")
				and AceLibrary("AceEvent-2.0") or false
		end
		if event then
			event:TriggerEvent("Links_LinkOpened", link, frame)
		end
		self:SendMessage("Links_LinkOpened", link, frame)
	end
end

local function MakeFrameIcon(frame)
	local icon = frame:CreateTexture(nil, "ARTWORK")
	frame.icon = icon
	icon:SetHeight(36)
	icon:SetWidth(36)
	icon:ClearAllPoints();
	icon:SetPoint("TOPRIGHT", frame, "TOPLEFT", "1", "-2")
	return icon
end

function Links:OpenLink(link, text, frame)
	link = link or (text and text:match("|H([^|]+)|h"))
	if not link then return end

	frame = frame or self:GetAvailableFrame()
	frame:ClearLines()
	frame:SetOwner(UIParent, "ANCHOR_PRESERVE")
	frame:SetHyperlink(link)
	frame.link = link
	frame.text = text

	local icon = frame.icon
	local p = self.db.profile
	if p.showicon then
		local type = strsplit(":", link, 2)
		if type == "item"  then
			icon = icon or MakeFrameIcon(frame)
			icon:SetTexture(GetItemIcon(link))
			icon:Show()
		elseif type == "spell" then
			icon = icon or MakeFrameIcon(frame)
			icon:SetTexture((select(3, GetSpellInfo(link:match("%d+")))))
			icon:Show()
		elseif type == "achievement" then
			icon = icon or MakeFrameIcon(frame)
			icon:SetTexture((select(10, GetAchievementInfo(link:match("%d+")))))
			icon:Show()
		elseif icon then
			icon:Hide()
		end
	elseif icon then
		icon:Hide()
	end

	self:InsertInLinksHistory(link, text)
	self:TriggerEvent(link, frame)
	frame:Show()
end

local function OpenHistoryLink(_, link, text)
	SetItemRef(link, text)
end

function Links:InsertInLinksHistory(link, text)
	local p = self.db.profile
	if p.history then
		local h = self.linkHistory
		if not h then
			h = {}
			self.linkHistory = h
		end
		local prev
		for i, entry in ipairs(h) do
			if entry.arg1 == link then
				prev = table.remove(h, i)
				break
			end
		end
		if not prev then
			prev = h[p.maxhistory] or { func = OpenHistoryLink }
			prev.text = text
			prev.arg1 = link
			prev.arg2 = text
		end
		table.insert(h, 1, prev)
		h[p.maxhistory + 1] = nil
	end
end

function Links:SetItemRef(link, text, ...)
	local type = strsplit(":", link, 2)
	if not self.db.profile.types[type] then
		return self.hooks.SetItemRef(link, text, ...)
	end
	if not self.db.profile.modinvert == not not IsModifiedClick("Links_Disable") then
		if not HandleModifiedItemClick(text) then
			ShowUIPanel(ItemRefTooltip)
			if not ItemRefTooltip:IsShown() then
				ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
			end
			ItemRefTooltip:SetHyperlink(link)
		end
	elseif not HandleModifiedItemClick(text) then
		for frame in pairs(self.frames) do
			if frame:IsVisible() and frame.link == link then
				frame:Hide()
				frame.link = nil
				return
			end
		end
		self:OpenLink(link, text)
	end
end

_G.Links = Links
